/* 
  EasyDeal Distributor: A machine which assigns 52 bridge playing cards to four bins (N,E,S,W)
  
  Two stepper motors are used:

      pickup =  a motor which extracts the bottom card of a pile
                and then retracts for a short moment to make sure that the next card
                keeps in position (it might have been moved some millimeters 
                together with the bottom card)
                
      select =  a motor which selects the desired bin by turning the desired
                catching bin towards the outlet of the card pile.
                bins are numbered from N = 0 to W = 3.
                works in 1/8 steps

  All steppers have 200 steps per cycle.
  
  The transport wheels for pickup and push have a diameter of 49.6 mm; therefore one step
  moves a card by 0.7775 mm.
  
  A card has a width of 56 mm and a length of 87 mm.

  We do not need an end switch. The user is expected to position the NORTH BIN 
  correctly before switching POWER ON.
  
  The hardware is a CNC V3 shield with NEMA 17 stepper motors controlled by A4988 driver modules.
  18 Volts and low current (~50 mA per motor) should be sufficient for all motors.
  The current for the bin select motor could be adjusted a little higher (~75mA).
  
  We use the AccelStepper library which allows to operate several steppers in prallel,
  using acceleration and deceleration.

  Typically, after 600 msec the next bottom card is in position to be scanned.
  Typically after 900 msec a whole cycle is completed.

  SEE ALSO   https://www.tinkercad.com/things/dt2yM2QFRWQ-mighty-blorr/editel?tenant=circuits

*/

#include <AccelStepper.h>

// ===============================================================================================

class Distributor {

	// the application class

	public:

		Distributor();

        void msg(String text);
        void enableMotors();
        void disableMotors();
        bool findZero();
        void unload();
        
        void dealSingleCard(int bin);
        void dealSingleCard(int bin, bool eject, bool retract);

        void dealDeckHands();
        void dealDeckCircular();
        void dealDeckRandom();
        void dealTest();
        
        int  buttonPressed();
        void handleCmd(int cmd);
		
        void setup();
		void loop();

	private:

		int   debug             =      1;   // 0 = nothing, 1 = normal, 2 = verbose

        int   prefetch          =   12*8;   // steps before bin arrival to start pickup
		
		// pickup
		int   pickupStepPin     =       2;  // aka X-AXIS step
		int   pickupDirPin      =       5;  // aka X-AXIS dir
		long  pickupSteps       =  -165*4;  // steps needed to pick and eject a card 
		long  pickupStepsBack   =    45*4;  // steps backwards to retract next card
		float pickupSpeedMax    = 10000./2.;  // never reached because the run time is too short
		float pickupAccelStart  = 60000./2.;  // steps / sec / sec
		float pickupAccelStop   =160000.0;  // steps / sec / sec
		AccelStepper  pickup;

		// select 
		int   selectStepPin      =      4;  // aka Z-AXIS step
		int   selectDirPin       =      7;  // aka Z-Axis dir
		float selectSpeedMax     = 4800.0;  // steps / second
		float selectAccel        = 6400.0;  // steps / second / second
		AccelStepper  select;

        // digital outputs
        int   motorEnablePin    =      8;   // low: enables motors, high: disables them

        // digital inputs
        int   greenButtonPin     =     11;  // aka Z+             - deal and scan   4 * 13 cards
        int   yellowButtonPin    =     10;  // aka Y+             - deal and reead 13 * 4  cards (circular)
        int   blueButtonPin      =      9;  // aka X+             - deal and read  52 * 1  cards (random)
        int   blackButtonPin     =     12;  // aka spindle enable - move select motor to zero position
        int   forkSensorPin      =     13;  // aka spindle dir    - detect zero position of select motor

		// relative movements from bin to bin
		int path[4][4] = { {0,1,2,-1},{-1,0,1,-2},{-2,-1,0,1},{1,2,-1,0} };
			
		// symbolic constants for states
		// =============================
		
		#define UNDEFINED           99
		
		// card state ( ~ machine state )
		int   cardState = UNDEFINED;
		#define CARD_PICKING        1         // pickup motor fwd, push motor fwd in sync
		#define CARD_RETRACTING     2         // pickup motor bwd, push motor still fwd accelerating
		#define CARD_READY_FOR_DEAL 9         // full cycle completed

		// bin selection state
		int   binState = UNDEFINED;
		#define BIN_MOVING          11        // select motor running towards destination
		#define BIN_ARRIVED         12        // bin has target position, select motor stopped
		#define BIN_READY_FOR_MOVE  19        // full cycle completed (card is in bin)

		// messages for the master (Raspi)
		#define MSG_READY_FOR_SCAN  "SCAN"

		// working variables to keep track of current bin and of processed cards
		int  selectPos,  lastBin;
		int  cardNumber   = 0;

		// ---- functions -----
		
		void showMotorState(AccelStepper,char *);
};

// ================================================================  CONSTRUCTOR
Distributor::Distributor() {
	// constructor: create stepper motors and assign digital IOs for control
    
	pickup  = AccelStepper(AccelStepper::DRIVER, pickupStepPin, pickupDirPin );
	select  = AccelStepper(AccelStepper::DRIVER, selectStepPin, selectDirPin );
};

// ================================================================  MSG
void Distributor::msg(String text) {
    Serial.println(text);
};

// ================================================================  ENABLE MOTORS
void Distributor::enableMotors() {
    digitalWrite(motorEnablePin, LOW);    
}

// ================================================================  DISABLE MOTORS
void Distributor::disableMotors() {
    digitalWrite(motorEnablePin, HIGH);    
}
// ================================================================  SHOW MOTOR STATE
void Distributor::showMotorState(AccelStepper motor,char *name) {
	Serial.print(name);Serial.print(": pos=");
	Serial.print(motor.currentPosition());
	Serial.print(" target=");
	Serial.print(motor.targetPosition());
	Serial.print(" distToGo=");
	Serial.print(motor.distanceToGo());
	Serial.print(" speed=");
	Serial.print(motor.speed());
	Serial.print(" running=");
	Serial.print(motor.isRunning());
	Serial.println();
}

// ================================================================  FIND ZERO
boolean Distributor::findZero() {
    // find the black spot, move beyond and then back into it        
    
    enableMotors();

    // define current position as 0
    select.setCurrentPosition(0);

    // if the sensor is covered: move away CW to get before the black spot (signal will be LOW)
    if (digitalRead(forkSensorPin)==HIGH) select.runToNewPosition(120);

    // rotate CCW to black spot (when we see a HIGH signal the LED got covered)
    select.moveTo(-1600);    // not more than a full circle
    for(;;) {
        if (digitalRead(forkSensorPin)==HIGH) break;
        if (!select.run()) break;
    }

    // turn further CCW so that we get behind the black spot (signal will be LOW)
    select.runToNewPosition(select.currentPosition()-120);

    // now turn CW until we reach the black spot again (=HIGH signal)
    select.moveTo(select.currentPosition()+200); // should not be very far
    int zeroPos= 9999;
    for(;;) {
        if (!select.run()) break;
        if (digitalRead(forkSensorPin) == HIGH) {
            zeroPos=select.currentPosition();
            break;
        }
    }

    if (zeroPos==9999) {
        // something went wrong
        disableMotors();
        return false;
    }
    
    // add the physical offset between switch and NORTH bin, positive= CW
    zeroPos += 280;
    select.runToNewPosition(zeroPos);
    select.setCurrentPosition(0);
    // Serial.println("zero = "+String(zeroPos));
    disableMotors();
    lastBin=0;
    selectPos=0;
}

// ================================================================  UNLOAD
void Distributor::unload() {
    // move bin to unload position (NORTH facing right front)
    
    msg("moving to UNLOAD position");
    enableMotors();
    select.runToNewPosition(200); // 45 degrees 
    lastBin=-1;
    delay(1000);
    disableMotors();
}

// ================================================================  DEAL SINGLE CARD
void Distributor::dealSingleCard(int bin) {
    dealSingleCard(bin,true,true);
}

// ================================================================  DEAL SINGLE CARD
void Distributor::dealSingleCard(int bin, bool eject, bool retract) {
    // process bottom card of pile: move bin into position,
    // pickup card and retract next card

    msg("dealing card "+String(cardNumber)+" -> #"+String(bin));

    if (eject) cardNumber++;    

	if (lastBin<0) findZero();	// if the selector has an undefined position
    if (lastBin==bin) {
        // bin does not need to be moved
        delay(50);
        binState = BIN_ARRIVED;
    }
    else {
        // set target position for bin by adding relative path to current position
        // 400 steps ~ 90 degrees
        selectPos+= -400 * path[lastBin][bin];
        select.moveTo(selectPos);
        binState = BIN_MOVING;
    }

    enableMotors();

    for (;;) {
        // move bin if necessary
        if (binState==BIN_MOVING && !select.run()) {
            // bin has reached its target position
            // normalize position
            selectPos %= 400;
            select.setCurrentPosition(selectPos);
            lastBin = bin;
            binState = BIN_ARRIVED;
        }
    
        // trigger pickup shortly before bin arrives
        if (cardState==CARD_READY_FOR_DEAL && (binState==BIN_ARRIVED || abs(select.distanceToGo())<prefetch)) {
            if (eject) {
                pickup.moveTo(pickupSteps);
                cardState = CARD_PICKING;
            }
            else if (retract) {
                pickup.move(floor(pickupStepsBack*0.5));
                cardState = CARD_RETRACTING;
            }
            else {
                break;
            }
        }
    
        // picking card ..
        if (cardState==CARD_PICKING && !pickup.run()) {
            if (retract) {
                // pickup finished, trigger retraction of next card
               	pickup.setAcceleration(pickupAccelStop);
                pickup.move(pickupStepsBack);
                cardState=CARD_RETRACTING;
            }
            else {
                break;
            }
        }
    
        // retracting next card ..
        else if (cardState==CARD_RETRACTING && !pickup.run()) {
            // retraction finished
           	pickup.setAcceleration(pickupAccelStart);
            break;
        }
    }

    // reset position
    pickup.setCurrentPosition(0);
    cardState=CARD_READY_FOR_DEAL;
    binState=BIN_READY_FOR_MOVE;

    disableMotors();

    // tell raspi that bottom card can now be scanned
    Serial.println(MSG_READY_FOR_SCAN); Serial.flush();
}

// ================================================================  DEAL DECK HANDS
void Distributor::dealDeckHands() {
    // assign the cards of each hand blockwise (13 cards each)

    msg("dealing 4 hands as blocks of 13 cards");
    
    findZero();
    cardNumber=0;
    for (int c=0;c<52;c++) {
        dealSingleCard(c/13);
    }
    unload();
}

// ================================================================  DEAL DECK CIRCULAR
void Distributor::dealDeckCircular() {
    // assign cards to bins one by one

    msg("dealing 52 cards 1 by 1");

    findZero();
    cardNumber=0;
    for (int c=0;c<52;c++) {
        dealSingleCard(c%4);
    }
    unload();
}

// ================================================================  DEAL DECK RANDOM
void Distributor::dealDeckRandom() {
    // assign cards to bins randomly, making sure every bin gets 13 cards
    
    msg("dealing 52 cards by random");

    findZero();
    cardNumber=0;

	// create a list of 13 * 4 bin numbers
    int bins[52];
    for (int c=0;c<52;c++) bins[c]=c%4;
	// shuffle bins by swapping random pairs
    for (int n=0;n<1000;n++) {
    	int b1=random(52);
    	int b2=random(52);
    	int b = bins[b1];
    	bins[b1]=bins[b2];
    	bins[b2]=b;
    }
    // deal one by one
    for (int c=0;c<52;c++) {
    	if (c>0 && bins[c]!=bins[c-1]) delay(100);	// allow card to drop before we move away
        dealSingleCard(bins[c]);
    }
    unload();
}

// ================================================================  DEAL TEST
void Distributor::dealTest() {
    // assign 4 cards to bins randomly, making sure every bin gets 1 card
    
    msg("dealing 4 cards by random");

    int kum[4] = { 0,0,0,0 };
    int bin;
    cardNumber=0;
    for (int c=0;c<4;c++) {
        for (;;) {
            bin=random(4);
            if (kum[bin]<1) break;
        }
        ++kum[bin];
        dealSingleCard(bin);
    }
}

// ================================================================  CHECK BUTTONS

int Distributor::buttonPressed() {
    if (digitalRead(greenButtonPin ) == LOW) return 'h';
    if (digitalRead(yellowButtonPin) == LOW) return 'c';
    if (digitalRead(blueButtonPin  ) == LOW) return 'r';
    if (digitalRead(blackButtonPin ) == LOW) return 'z';
    return '-';
}

// ================================================================  HANDLE CMD

void Distributor::handleCmd(int cmd) {
    // perform selected action

    msg("handle cmd "+String((char)(cmd)));
    
    if      (cmd=='h') {        // 4 * 13 circular
        dealDeckHands();
    }
    else if (cmd=='c') {        // 52 * 1 circular
        dealDeckCircular();
    }
    else if (cmd=='r') {        // 52 * 1..x random
        dealDeckRandom();
    }
    else if (cmd=='s') {
        dealSingleCard(lastBin);            // 1 card to same bin
    } 
    else if (cmd=='e') {
        msg("eject card");
        dealSingleCard(lastBin,true,false); // 1 card eject only
    }
    else if (cmd=='w') {
        msg("withdraw (retract) card");
        dealSingleCard(lastBin,false,true); // 1 card withdraw (short amount) only
    }
    else if (cmd=='t') {        // 4 * 1 random
        dealTest();
    }
    else if (cmd=='u') {        // unload position
        unload();
    }
    else if (cmd=='z') {        // find zero position
        msg("going to zero position");
        findZero();
    }
    else if (cmd=='N' || cmd=='1') {                // single card to North
        dealSingleCard(0);
    }
    else if (cmd=='E' || cmd=='O' || cmd=='2') {    // single card to East
        dealSingleCard(1);
    }
    else if (cmd=='S' || cmd=='3') {                // single card to South
        dealSingleCard(2);
    }
    else if (cmd=='W' || cmd=='4') {                // single card to West
        dealSingleCard(3);
    }
    else if (cmd=='P' || cmd=='p') {                // change pickup AND retract distance
        float factor = (cmd=='p') ? 0.9 : 1./0.9;
        int old = pickupSteps;
        pickupSteps = round(old * factor);
        Serial.println("changing PICKUP  steps from "+String(old)+" to "+String(pickupSteps));
        old=pickupStepsBack;
        pickupStepsBack = round(old * factor);
        Serial.println("changing RETRACT steps from "+String(old)+" to "+String(pickupStepsBack));
        findZero(); // move bin rotor for physical confirmation        
    }
    else if (cmd=='Q' || cmd=='q') {                // change RETRACT distance only
        float factor = (cmd=='q') ? 0.9 : 1./0.9;
        int old=pickupStepsBack;
        pickupStepsBack = round(old * factor);
        Serial.println("changing RETRACT steps from "+String(old)+" to "+String(pickupStepsBack));
        findZero(); // move bin rotor for physical confirmation        
    }
    else {
        // ignore all other cmds
    }
}

// ================================================================  SETUP
void Distributor::setup() {

	// use serial interface to talk to development PC or Raspi
	Serial.begin(115200);
    delay(100);
    
	if (debug>0) Serial.println("EasyDeal Distributor starting ..");

    // configure GP INPUTS for fork sensor and control buttons    
    pinMode(forkSensorPin,  INPUT       );  // zero switch for select motor
    pinMode(greenButtonPin, INPUT_PULLUP);  // deal 4 * 13 hands
    pinMode(yellowButtonPin,INPUT_PULLUP);  // deal 13 * 4 circular
    pinMode(blueButtonPin,  INPUT_PULLUP);  // deal 52 * 1 random
    pinMode(blackButtonPin, INPUT_PULLUP);  // move selector to zero

    // configure GP OUTPUTS for motor power    
    pinMode(motorEnablePin, OUTPUT      );  // power enable pin

	// configure motors
	pickup.setMaxSpeed(pickupSpeedMax);
	pickup.setAcceleration(pickupAccelStart);
	select.setMaxSpeed(selectSpeedMax);
	select.setAcceleration(selectAccel);
	
    // find initial position of selector (NORTH facing the card outlet)
    findZero();

    // set initial states of card feeder and bin selector
    cardState  = CARD_READY_FOR_DEAL;
    binState   = BIN_READY_FOR_MOVE;
    selectPos=0;

    // show initial motor states
    if (debug>=2) showMotorState(select ,"select is at bin #0");
    if (debug>=2) showMotorState(pickup,"pickup is ready");

	// tell raspi that we are ready to deal
    Serial.println(MSG_READY_FOR_SCAN);	Serial.flush();
}

// ================================================================  LOOP

void Distributor::loop() {
    // fetch next command from serial line
  	
    delay(1);
    // check physical buttons
    int c = buttonPressed();

    // check serial input
    if (c== '-') {
        if (!Serial.available()) return;        // nothing from host
        c = Serial.read();                      // cmd from host
        if (c==0 || c==10 | c==13) return;      // ignore these chars
    }
    
    // execute command
    msg("cmd="+String((char)(c)));
	handleCmd(c);
}

// ===================================================================================================

// create EasyDeal Distributor instance and use it for setup() and loop()

Distributor distributor = Distributor();
void setup() { distributor.setup(); }
void loop()  { distributor.loop();  }
