
const fs 			= require("fs");
const jpeg 			= require('jpeg-js');
const UTIF 			= require('utif');
const spawnProcess	= require('child_process').spawn;
const { Image } 	= require('image-js');

class ImageLoader {
	
	// Initially we once spawn a ffmpeg task which will connect to the camera.
	// ffmpeg will crop the relevant image region and stream 32 bit RGBA tuplets.
	// We connect to the output pipe of the task and keep permanently
	// collecting chunks until we have a complete frame of ~ 100 kB.
	// When we receive a chunk of exactly 64kB we assume that this is only
	// some part of a frame. If the chunk is shorter we assume that this
	// is the final chunk of a frame.
	// If there was no request to startRecognition() we will simply
	// throw away the completed frame and restart collecting chunks.
	// If there is a request to startRecognition() we will start detection
	// with the next frame. We convert the rgba32 pixels to (uncompressed)
	// tif format, load the tif data into the image-js library 
	// for further transformation and apply our detection logic. 
	// As soon as detection is ready we create a compressed jpg image
	// from the frame data (rgba buffer) and inform the connected web clients
	// that they can pick up the image.
	
	constructor(width,height,left,top) {
		// the crop area where we can find the combined symbols for rank and suit
		this.width	= width;
		this.height	= height;
		this.left	= left;
		this.top	= top;
		this.pixels = [];				// a binary matrix (height x width)
										// true = black pixel, false = white pixel
		this.receiveScanResult = null;	// setting this to a callback function start scanning
										// and finally invokes the callback function
										// the value is automaticall reset to null
										// after a scan has been conducted
	}	
	
	log(msg) {
		console.log(Date.now()%100000+"\tImageLoader     "+msg);
	}
	
	startRecognition(callback) {
		// starts recognition for the next frame and then invokes the callback
		// immediately returns to the caller
		this.receiveScanResult=callback;
		this.log("recognition requested");
	}
	
	collectFrames(cmdLine) {
		// spawn the streaming process; it will do the cropping and it will
		// deliver frames at a decent rate of (e.g.) 5 frames per second
		this.log("starting ffmpeg stream "+cmdLine);
		var ffmpeg = spawnProcess('ffmpeg', cmdLine.split(" "));
		
		var	rgbas = Buffer.alloc(0);	// buffer with pixel triplets

		var that = this;
		// action to be taken on reception of an RGBA chunk
		ffmpeg.stdout.on('data', async function (data) {
			try {
				// this.log("receiving data "+data.length);
				if (data.length==65536) {
					// if we get a full buffer we assume that the frame is not yet complete
					// so we just keep collecting data
					rgbas = Buffer.concat([rgbas,data]);
				}
				else {
					// we just got the last chunk of a frame;
					
					// check if we shall start analysing the image
					if (that.receiveScanResult==null) {
						// reset buffer and return
						rgbas=Buffer.alloc(0);
						// that.log("skipping frame")
						return;
					}
										
					// concatenate current data chunk to complete the frame
					rgbas = Buffer.concat([rgbas,data]);					
					
					var tifImage = UTIF.encodeImage(rgbas,that.width,that.height);
					// that.log("tif ready ",tifImage);

					// create image-js instance
					let image = await Image.load(tifImage);

					// perform detection
					var result = new Detector(image).detect();
					
					// invoke the callback to tell the result
					that.receiveScanResult(result);
					
					// produce a jpeg version of the image so it can be shown by the clients´browsers
					that.log("encoding raw "+rgbas.length+" bytes ,w="+that.width+", h="+that.height);
					var jpgImage = jpeg.encode({data:rgbas,width:that.width,height:that.height},80);
					that.log("jpg ready ",jpgImage);

					// publish jpeg image
					
					
					// reset callback (must be retriggered by caller to start the next recognition)
					that.receiveScanResult=null;
					
					// reset the buffer, so we can catch the next frame
					rgbas=Buffer.alloc(0);			// reset buffer for next scan
				}
			} catch(e) {
				console.log(e);
			} 
		});

		ffmpeg.on('close', function (code) {
			that.log('ffmpeg process terminated, rc=' + code);
		});
	}
	

	start(maxFrames) {
		// this.log("collecting "+maxFrames+" frames...");
		
		// let us collect maxFrames or forever if maxFrames<0
		// at a rate of 5 frames / second
		this.collectFrames(
			"-i /dev/video0"+(maxFrames>=0 ? " -vframes "+maxFrames : "")
			+ " -filter:v crop="+this.width+":"+this.height+":"+this.left+":"+this.top
			+" -r 5 -f rawvideo -pix_fmt rgb32 -s "+this.width+"x"+this.height+" pipe:1"
		);
	}
}

class Detector {
	// expects an image-js object and offers recognition algorithms
	
	constructor(image) {
		// reduce color depth and get red channel
		this.img = image
			.colorDepth(8)
			.getChannel(1)
			.flipX()
		;
	}
	
	log(msg) {
		console.log(Date.now()%100000+"Detector        "+msg);
	}
	
	detect() {
		
		// calculate black threshold
		var black = this.img.getThreshold({algorithm:"otsu"});	
		this.log("scanning image "+this.img.width+" x "+this.img.height+" , black = "+black);

		// get pixel array, convert to binary matrix using black threshold
		var pix = [];
		for (var i=0, y=-1; i < this.img.data.length; i++) {
			if (i % this.img.width === 0) pix[++y] = [];
			pix[y].push(this.img.data[i]<=black ? 1 : 0);
		}
		
		// remove some speckles which could jeopardize finding the correct image doimensions for rank and suit
		for (var x = 0; x < this.img.width-2;x++) {
			for (var y = 1; y<this.img.height-1; y++) {
				if ((pix[y][x] && !pix[y-1][x] && !pix[y+1][x] && (pix[y-1][x+1]+pix[y][x+1]+pix[y+1][x+1]) <= 2) 
				 || (pix[y][x] && (pix[y-1][x] + pix[y+1][x]<=1) && (pix[y-1][x+1]+pix[y][x+1]+pix[y+1][x+1]) <= 1)) {
					pix[y][x]=0;
					// this.log("despeckle "+x+" "+y);
				}
			}
		}

		// this.showMatrix(pix);
		
		return ("done with some result");

	}
	
	showMatrix(data) {
		for (var y=0; y<data.length; y++) {
			var line=(""+y).padStart(3)+" ";
			for (var x=0; x<data[0].length; x++) {
				line=line+(data[y][x] ? "x":".");
			}
			console.log(line);
		}
	}
	
}


// ==========================================================

// setup ImageLoader
var IL = new ImageLoader(110,290,120,100);
IL.start(-1);	// collect N frames

// ask regularly for a scan
setInterval( function() {
		IL.startRecognition(receiveScanResult);
	},1000
);

function receiveScanResult(result) {
	console.log(Date.now()%100000 + "\t"+result);
}
